# -*- coding: utf-8 -*-

"""Data structures for Zenodo."""

import datetime
from typing import Optional, Sequence

from pydantic import BaseModel, Field
from typing_extensions import Literal

__all__ = [
    "Creator",
    "Metadata",
]


# https://developers.zenodo.org/#rest-api


class Creator(BaseModel):
    """A creator, see https://developers.zenodo.org/#representation."""

    name: str = Field(
        ..., description="Name of the creator in the format Family name, given names", example="Hoyt, Charles Tapley"
    )
    affiliation: Optional[str] = Field(description="affiliation of the creator", example="Harvard Medical School")
    orcid: Optional[str] = Field(description="ORCID identifier of the creator", example="0000-0003-4423-4370")
    gnd: Optional[str] = Field(
        description="German National Library identifier of the creator. "
        "See also https://www.wikidata.org/wiki/Property:P227."
    )

    @property
    def orcid_url(self) -> Optional[str]:
        """Get the ORCID identifier as a URL."""
        return f"https://orcid.org/{self.orcid}" if self.orcid else None

    @property
    def gnd_url(self) -> Optional[str]:
        """Get the GND identifier as a URL."""
        return f"https://d-nb.info/gnd/{self.gnd}"

    def __post_init__(self):  # noqa:D105
        if "," not in self.name:
            raise ValueError("name should be in format Family name, given names")


UploadType = Literal[
    "publication",
    "poster",
    "presentation",
    "dataset",
    "image",
    "video",
    "software",
    "lesson",
    "physicalobject",
    "other",
]

PublicationType = Literal[
    "annotationcollection",
    "book",
    "section",
    "conferencepaper",
    "datamanagementplan",
    "article",
    "patent",
    "preprint",
    "deliverable",
    "milestone",
    "proposal",
    "report",
    "softwaredocumentation",
    "taxonomictreatment",
    "technicalnote",
    "thesis",
    "workingpaper",
    "other",
]

ImageType = Literal[
    "figure",
    "plot",
    "drawing",
    "diagram",
    "photo",
    "other",
]
AccessRight = Literal[
    "open",
    "embargoed",
    "restricted",
    "closed",
]


def _today_str() -> str:
    return datetime.datetime.today().strftime("%Y-%m-%d")


class Metadata(BaseModel):
    """Metadata for the Zenodo deposition API."""

    title: str
    upload_type: UploadType
    description: str
    creators: Sequence[Creator]
    access_right: AccessRight = "open"
    language: Optional[str] = "eng"
    version: Optional[str] = Field(default_factory=_today_str)
    license: Optional[str] = "CC0-1.0"
    publication_type: Optional[PublicationType] = None
    image_type: Optional[ImageType] = None

    def __post_init__(self):  # noqa:D105
        if self.upload_type == "publication":
            if self.publication_type is None:
                raise ValueError("missing publication_type")
        elif self.upload_type == "image":
            if self.image_type is None:
                raise ValueError("missing image_type")
        if self.access_right in {"open", "embargoed"}:
            if self.license is None:
                raise ValueError(f"need a license for access_right={self.access_right}")



# -*- coding: utf-8 -*-

"""A wrapper for the Zenodo API."""

from .api import (  # noqa:F401
    Zenodo,
    create_zenodo,
    download_zenodo,
    download_zenodo_latest,
    ensure_zenodo,
    update_zenodo,
)
from .struct import Creator, Metadata  # noqa:F401



"""Tests for the upload and revision lifecyle."""

import logging
import tempfile
import unittest
from pathlib import Path
from uuid import uuid4

import pystow

from zenodo_client import Creator, Metadata, Zenodo

logger = logging.getLogger(__name__)

TEXT_V1 = "this is some test text\noh yeah baby"
TEXT_V2 = "this is some v2 test text\noh yeah baby"
TEXT_V3 = "this is some v3 test text\noh yeah baby"

ACCESS_TOKEN = pystow.get_config("zenodo", "sandbox_api_token")


@unittest.skipUnless(ACCESS_TOKEN, reason="Missing Zenodo sandbox API token")
class TestLifecycle(unittest.TestCase):
    """Test case for zenodo client."""

    def setUp(self) -> None:
        """Set up the test case with a zenodo client."""
        self.zenodo = Zenodo(sandbox=True, access_token=ACCESS_TOKEN)
        self.key = f"test-{uuid4()}"
        self._directory = tempfile.TemporaryDirectory()
        self.directory = Path(self._directory.name).resolve()

    def tearDown(self) -> None:
        """Tear down the test case."""
        self._directory.cleanup()

    def test_create(self):
        """Test creating a record."""
        data = Metadata(
            title="Test Upload",
            upload_type="dataset",
            description="test description",
            creators=[
                Creator(
                    name="Hoyt, Charles Tapley",
                    affiliation="Harvard Medical School",
                    orcid="0000-0003-4423-4370",
                ),
            ],
        )
        path = self.directory.joinpath("test.txt")
        path.write_text(TEXT_V1)

        logger.info(f"wrote data to {path}")
        res = self.zenodo.ensure(
            key=self.key,
            data=data,
            paths=path,
        )
        res_json = res.json()
        # print(f"\n\nSEE V1 ON ZENODO: {res_json['links']['record_html']}")

        self.assertEqual(True, res_json["submitted"])
        self.assertEqual("done", res_json["state"])
        self.assertEqual("dataset", res_json["metadata"]["upload_type"])
        self.assertEqual(data.title, res_json["metadata"]["title"])
        self.assertEqual(data.version, res_json["metadata"]["version"])

        deposition_id = res_json["record_id"]
        # print(f"DEPOSITION ID: {deposition_id}")
        path.write_text(TEXT_V2)

        res = self.zenodo.update(deposition_id, paths=path)
        res_json = res.json()
        # print(f"SEE V2 ON ZENODO: {res_json['links']['record_html']}")
        self.assertEqual(f"{data.version}-1", res_json["metadata"]["version"])

        path.write_text(TEXT_V3)
        res = self.zenodo.update(deposition_id, paths=path)
        res_json = res.json()
        # print(f"SEE V3 ON ZENODO: {res_json['links']['record_html']}")
        self.assertEqual(f"{data.version}-2", res_json["metadata"]["version"])

    def test_create_no_orcid(self):
        """Test create with no ORCID."""
        data = Metadata(
            title="Test Upload",
            upload_type="dataset",
            description="test description",
            creators=[
                Creator(
                    name="Hoyt, Charles Tapley",
                    affiliation="Harvard Medical School",
                ),
            ],
        )
        path = self.directory.joinpath("test.txt")
        path.write_text(TEXT_V1)

        logger.info(f"wrote data to {path}")
        res = self.zenodo.ensure(
            key=self.key,
            data=data,
            paths=path,
        )
        res_json = res.json()
        # print(f"\n\nSEE V1 ON ZENODO: {res_json['links']['record_html']}")

        self.assertEqual(True, res_json["submitted"])
        self.assertEqual("done", res_json["state"])
        self.assertEqual("dataset", res_json["metadata"]["upload_type"])
        self.assertEqual(data.title, res_json["metadata"]["title"])
        self.assertEqual(data.version, res_json["metadata"]["version"])



# -*- coding: utf-8 -*-

"""Command line interface for :mod:`zenodo_client`.

Why does this file exist, and why not put this in ``__main__``? You might be tempted to import things from ``__main__``
later, but that will cause problems--the code will get executed twice:

- When you run ``python3 -m zenodo_client`` python will execute``__main__.py`` as a script.
  That means there won't be any ``zenodo_client.__main__`` in ``sys.modules``.
- When you import __main__ it will get executed again (as a module) because
  there's no ``zenodo_client.__main__`` in ``sys.modules``.

.. seealso:: https://click.palletsprojects.com/en/7.x/setuptools/#setuptools-integration
"""

import logging

import click
from more_click import verbose_option

from .api import download_zenodo, download_zenodo_latest, update_zenodo

__all__ = ["main"]

logger = logging.getLogger(__name__)


@click.group()
def main():
    """CLI for Zenodo Client."""


@main.command()
@click.argument("deposition")
@click.argument("path")
@click.option("--force", is_flag=True)
@click.option("--latest", is_flag=True)
def download(deposition: str, path: str, force: bool, latest: bool):
    """Ensure a record is downloaded."""
    if latest:
        download_zenodo_latest(deposition, path, force=force)
    else:
        download_zenodo(deposition, path, force=force)


@main.group()
@click.argument("deposition")
@click.argument("paths", nargs=-1)
@verbose_option
@click.version_option()
def update(deposition, paths):
    """Update the record and given files."""
    update_zenodo(deposition, paths)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""A client for Zenodo."""

import datetime
import logging
import os
import time
from pathlib import Path
from typing import Any, Callable, Iterable, List, Mapping, Optional, Sequence, Union

import pystow
import requests

from .struct import Metadata

__all__ = [
    "ensure_zenodo",
    "update_zenodo",
    "create_zenodo",
    "download_zenodo",
    "download_zenodo_latest",
    "Zenodo",
]

logger = logging.getLogger(__name__)

Data = Union[Mapping[str, Any], Metadata]

PartsFunc = Callable[[str, str, str], Sequence[str]]
PartsHint = Union[None, Sequence[str], PartsFunc]
Paths = Union[str, Path, Iterable[str], Iterable[Path]]


def ensure_zenodo(key: str, data: Data, paths: Paths, **kwargs) -> requests.Response:
    """Create a Zenodo record if it doesn't exist, or update one that does."""
    return Zenodo(**kwargs).ensure(key=key, data=data, paths=paths)


def create_zenodo(data: Data, paths: Paths, **kwargs) -> requests.Response:
    """Create a Zenodo record."""
    return Zenodo(**kwargs).create(data, paths)


def update_zenodo(deposition_id: str, paths: Paths, **kwargs) -> requests.Response:
    """Update a Zenodo record."""
    return Zenodo(**kwargs).update(deposition_id, paths)


def download_zenodo(deposition_id: str, name: str, force: bool = False, **kwargs) -> Path:
    """Download a Zenodo record."""
    return Zenodo(**kwargs).download(deposition_id, name=name, force=force)


def download_zenodo_latest(deposition_id: str, path: str, force: bool = False, **kwargs) -> Path:
    """Download the latest Zenodo record."""
    return Zenodo(**kwargs).download_latest(deposition_id, name=path, force=force)


class Zenodo:
    """A wrapper around parts of the Zenodo API."""

    def __init__(self, access_token: Optional[str] = None, sandbox: bool = False):
        """Initialize the Zenodo class.

        :param access_token: The Zenodo API. Read with :mod:`pystow` from zenodo/api_token
            of zenodo/sandbox_api_token if in sandbox mode.
        :param sandbox: If true, run in the Zenodo sandbox.
        """
        self.sandbox = sandbox
        if self.sandbox:
            self.base = "https://sandbox.zenodo.org"
            # Use subsection support introduced in PyStow in
            # https://github.com/cthoyt/pystow/pull/59
            self.module = "zenodo:sandbox"
            self.access_token = pystow.get_config(self.module, "api_token", passthrough=access_token)
            if self.access_token is None:
                # old-style fallback
                self.access_token = pystow.get_config("zenodo", "sandbox_api_token", raise_on_missing=True)
        else:
            self.base = "https://zenodo.org"
            self.module = "zenodo"
            self.access_token = pystow.get_config(
                self.module, "api_token", passthrough=access_token, raise_on_missing=True
            )

        # Base URL for the API
        self.api_base = self.base + "/api"
        logger.debug("using Zenodo API at %s", self.api_base)

        # Base URL for depositions, relative to the API base
        self.depositions_base = self.api_base + "/deposit/depositions"

    def ensure(self, key: str, data: Data, paths: Paths) -> requests.Response:
        """Create a Zenodo record if it doesn't exist, or update one that does."""
        deposition_id = pystow.get_config(self.module, key)
        if deposition_id is not None:
            logger.info("mapped local key %s to deposition %s", key, deposition_id)
            return self.update(deposition_id=deposition_id, paths=paths)

        res = self.create(data=data, paths=paths)
        # Write the ID to the key in the local configuration
        # so it doesn't need to be created from scratch next time
        pystow.write_config(self.module, key, str(res.json()["id"]))
        return res

    def create(self, data: Data, paths: Paths) -> requests.Response:
        """Create a record.

        :param data: The JSON data to send to the new data
        :param paths: Paths to local files to upload
        :return: The response JSON from the Zenodo API
        :raises ValueError: if the response is missing a "bucket"
        """
        if isinstance(data, Metadata):
            logger.debug("serializing metadata")
            data = {
                "metadata": {key: value for key, value in data.dict(exclude_none=True).items() if value},
            }

        res = requests.post(
            self.depositions_base,
            json=data,
            params={"access_token": self.access_token},
        )
        res.raise_for_status()

        res_json = res.json()
        bucket = res_json.get("links", {}).get("bucket")
        if bucket is None:
            raise ValueError(f"No bucket in response. Got: {res_json}")

        logger.info("uploading files to bucket %s", bucket)
        self._upload_files(bucket=bucket, paths=paths)

        deposition_id = res_json["id"]
        logger.info("publishing files to deposition %s", deposition_id)
        return self.publish(deposition_id)

    def publish(self, deposition_id: str, sleep: bool = True) -> requests.Response:
        """Publish a record that's in edit mode.

        :param deposition_id: The identifier of the deposition on Zenodo. It should be in edit mode.
        :param sleep: Sleep for one second just in case of race conditions. If you're feeling lucky and rushed, you
            might be able to get away with disabling this.
        :return: The response JSON from the Zenodo API
        """
        if sleep:
            time.sleep(1)
        res = requests.post(
            f"{self.depositions_base}/{deposition_id}/actions/publish",
            params={"access_token": self.access_token},
        )
        res.raise_for_status()
        return res

    def update(self, deposition_id: str, paths: Paths) -> requests.Response:
        """Create a new version of the given record with the given files."""
        # Prepare a new version based on the old version
        # see: https://developers.zenodo.org/#new-version)
        res = requests.post(
            f"{self.depositions_base}/{deposition_id}/actions/newversion",
            params={"access_token": self.access_token},
        )
        res.raise_for_status()

        # Parse out the new version (@zenodo please give this as its own field!)
        new_deposition_id = res.json()["links"]["latest_draft"].split("/")[-1]

        # Get all metadata associated with the new version (this has updated DOIs, etc.)
        # see: https://developers.zenodo.org/#retrieve
        res = requests.get(
            f"{self.depositions_base}/{new_deposition_id}",
            params={"access_token": self.access_token},
        )
        res.raise_for_status()
        new_deposition_data = res.json()
        # Update the version
        new_deposition_data["metadata"]["version"] = _prepare_new_version(new_deposition_data["metadata"]["version"])
        new_deposition_data["metadata"]["publication_date"] = datetime.datetime.today().strftime("%Y-%m-%d")

        # Update the deposition for the new version
        # see: https://developers.zenodo.org/#update
        res = requests.put(
            f"{self.depositions_base}/{new_deposition_id}",
            json=new_deposition_data,
            params={"access_token": self.access_token},
        )
        res.raise_for_status()

        bucket = new_deposition_data["links"]["bucket"]

        # Upload new files. It calculates the hash on all of these, and if no files have changed,
        #  there will be no update
        self._upload_files(bucket=bucket, paths=paths)

        # Send the publish command
        return self.publish(new_deposition_id)

    def _upload_files(self, *, bucket: str, paths: Paths) -> List[requests.Response]:
        _paths = [paths] if isinstance(paths, (str, Path)) else paths
        rv = []
        # see https://developers.zenodo.org/#quickstart-upload
        for path in _paths:
            with open(path, "rb") as file:
                res = requests.put(
                    f"{bucket}/{os.path.basename(path)}",
                    data=file,
                    params={"access_token": self.access_token},
                )

            res.raise_for_status()
            rv.append(res)
        return rv

    def get_record(self, record_id: Union[int, str]) -> requests.Response:
        """Get the metadata for a given record."""
        res = requests.get(
            f"{self.api_base}/records/{record_id}",
            params={"access_token": self.access_token},
        )
        res.raise_for_status()
        return res

    def get_latest_record(self, record_id: Union[int, str]) -> str:
        """Get the latest record related to the given record."""
        res_json = self.get_record(record_id).json()
        # Still works even in the case that the given record ID is the latest.
        latest = res_json["links"]["latest"].split("/")[-1]
        logger.debug("latest for zenodo.record:%s is zenodo.record:%s", record_id, latest)
        return latest

    def download(self, record_id: Union[int, str], name: str, *, force: bool = False, parts: PartsHint = None) -> Path:
        """Download the file for the given record.

        :param record_id: The Zenodo record id
        :param name: The name of the file in the Zenodo record
        :param parts: Optional arguments on where to store with :func:`pystow.ensure`. If none given, goes in
            ``<PYSTOW_HOME>/zendoo/<CONCEPT_RECORD_ID>/<RECORD>/<PATH>``. Where ``CONCEPT_RECORD_ID`` is the
            consistent concept record ID for all versions of the same record. If a function is given, the function
            should take 3 position arguments: concept record id, record id, and version, then return a sequence for
            PyStow. The name of the file is automatically appended to the end of the sequence.
        :param force: Should the file be re-downloaded if it already is cached? Defaults to false.
        :returns: the path to the downloaded file.
        :raises FileNotFoundError: If the Zenodo record doesn't have a file with the given name

        For example, to download the most recent version of NSoC-KG, you can
        use the following command:

        >>> path = Zenodo().download('4574555', 'triples.tsv')

        Even as new versions of the data are uploaded, this command will always
        be able to check if a new version is available, download it if it is, and
        return the local file path. If the most recent version is already downloaded,
        then it returns the local file path to the cached file.

        The file path uses :mod:`pystow` under the ``zenodo`` module and uses the
        "concept record ID" as a submodule since that is the consistent identifier
        between different records that are versions of the same data.
        """
        res_json = self.get_record(record_id).json()
        # conceptrecid is the consistent record ID for all versions of the same record
        concept_record_id = res_json["conceptrecid"]
        # FIXME send error report to zenodo about this - shouldn't version be required?
        version = res_json["metadata"].get("version", "v1")
        logger.debug("version for zenodo.record:%s is %s", record_id, version)

        for file in res_json["files"]:
            if file["key"] == name:
                url = file["links"]["self"]
                break
        else:
            raise FileNotFoundError(f"zenodo.record:{record_id} does not have a file with key {name}")

        if parts is None:
            parts = [self.module.replace(":", "-"), concept_record_id, version]
        elif callable(parts):
            parts = parts(concept_record_id, str(record_id), version)
        return pystow.ensure(*parts, name=name, url=url, force=force)

    def download_latest(
        self,
        record_id: Union[int, str],
        name: str,
        *,
        force: bool = False,
        parts: PartsHint = None,
    ) -> Path:
        """Download the latest version of the file."""
        latest_record_id = self.get_latest_record(record_id)
        return self.download(latest_record_id, name=name, force=force, parts=parts)


def _prepare_new_version(old_version: str) -> str:
    new_version = datetime.datetime.today().strftime("%Y-%m-%d")
    if old_version == new_version:
        new_version += "-1"
    elif old_version.startswith(new_version) and old_version[-2] == "-" and old_version[-1].isnumeric():
        new_version += "-" + str(1 + int(old_version[-1]))  # please don't do this more than 10 times a day
    return new_version
